// SPDX-License-Identifier: LGPL-2.1-or-later
// SPDX-FileCopyrightText: 2006 Mathias Froehlich <mathias.froehlich@web.de>

// 数学工具类
// 提供了一系列数学计算函数，包括最大值、最小值、无溢出加法、渐进、幂运算、归一化、四舍五入等

#ifndef SGMisc_H
#define SGMisc_H

#include <cmath>

template<typename T>
class SGMisc {
public:
  static T pi() { return T(3.1415926535897932384626433832795029L); }
  static T twopi() { return 2*T(3.1415926535897932384626433832795029L); }

  static T min(const T& a, const T& b)
  { return a < b ? a : b; }
  static T min(const T& a, const T& b, const T& c)
  { return min(min(a, b), c); }
  static T min(const T& a, const T& b, const T& c, const T& d)
  { return min(min(min(a, b), c), d); }
  static T max(const T& a, const T& b)
  { return a > b ? a : b; }
  static T max(const T& a, const T& b, const T& c)
  { return max(max(a, b), c); }
  static T max(const T& a, const T& b, const T& c, const T& d)
  { return max(max(max(a, b), c), d); }

  // clip the value of a to be in the range between and including _min and _max
  // 如果a小于_min，则返回_min，如果a大于_max，则返回_max，否则返回a
  static T clip(const T& a, const T& _min, const T& _max)
  { return max(_min, min(_max, a)); }


  /// Add two (integer) values taking care of overflows.
  // 无溢出加法
  // 如果b大于0，则检查a+b是否溢出，如果溢出则返回SGLimits<T>::max()，否则返回a+b
  // 如果b小于0，则检查a+b是否溢出，如果溢出则返回SGLimits<T>::min()，否则返回a+b
  static T addClipOverflow(T a, T b)
  {
    if( b > 0 )
    {
      if( SGLimits<T>::max() - b < a )
        return SGLimits<T>::max();
    }
    else
    {
      if( SGLimits<T>::min() - b > a )
        return SGLimits<T>::min();
    }

    return a + b;
  }

  /// Add two (integer) values in place, taking care of overflows.
  // 无溢出原位置加法
  static T& addClipOverflowInplace(T& a, T b)
  {
    return a = addClipOverflow(a, b);
  }

  /**
   * Seek a variable towards a target value with given rate and timestep
   *
   * @param var     Variable or eg. SGPropObj
   * @param target  Target value
   * @param rate    Max. change rate/sec
   * @param dt      Time step (sec)
   */
  // 渐进： var 将以指定的速率 rate 按时间 dt 的递增向 target 渐进。
  template<class Var>
  static T seek(Var& var, T target, T rate, T dt)
  {
    if( var < target )
      return var = min(var + rate * dt, target);
    else
      return var = max(var - rate * dt, target);
  }

  /**
   * Get @c base raised to the power of @c N
   *
   * @tparam N      Exponent
   * @param base    Base
   */
  // 幂运算
  template<int N>
  static T pow(T base)
  {
    return (N < 0)
      ? (1. / pow<-N>(base))
      : (  ((N & 1) ? base : 1)
        * ((N > 1) ? pow<N / 2>(base * base) : 1)
        );
  }

  static int sign(const T& a)
  {
    if (a < -SGLimits<T>::min())
      return -1;
    else if (SGLimits<T>::min() < a)
      return 1;
    else
      return 0;
  }

  static T rad2deg(const T& val)
  { return val*180/pi(); }
  static T deg2rad(const T& val)
  { return val*pi()/180; }

  // normalize the value to be in a range between [min, max]
  // 归一化周期值：将 value 归一化到 [min, max] 周期范围内
  static T
  normalizePeriodic(const T& min, const T& max, const T& value)
  {
    T range = max - min;
    if (range < SGLimits<T>::min())
      return min;
    T normalized = value - range*floor((value - min)/range);
    // two security checks that can only happen due to roundoff
    if (normalized <= min)
      return min;
    if (max <= normalized)
      return min;
    return normalized;
  }

  // normalize the angle to be in a range between [-pi, pi]
  // 归一化角度：将 angle 归一化到 [-pi, pi] 范围内
  static T
  normalizeAngle(const T& angle)
  { return normalizePeriodic(-pi(), pi(), angle); }

  // normalize the angle to be in a range between [0, 2pi]
  // 归一化角度：将 angle 归一化到 [0, 2pi] 范围内
  static T
  normalizeAngle2(const T& angle)
  { return normalizePeriodic(0, twopi(), angle); }

  // 四舍五入：将 v 四舍五入到最接近的整数，返回结果为 T 类型
  static T round(const T& v)
  { return floor(v + T(0.5)); }

  // 将 v 四舍五入到最接近的整数
  static int roundToInt(const T& v)
  { return int(round(v)); }

  // Linear interpolation between two arbitrary typed values
  // 线性插值：在 val0 和 val1 之间进行线性插值，t 是插值因子，取值范围为 [0, 1]
  template<typename S>
  static S lerp(const S& val0, const S& val1, const T& t)
  { return val0*(T(1) - t) + val1*t; }

  /// Returns true if v is a NaN value
  /// Use with care: allways code that you do not need to use that!
  static bool isNaN(const T& v)
  {
    return std::isnan(v);
  }

  // 判断两个值是否相等
  static bool eq(const T& a, const T& b, const T& epsilon = SGLimits<T>::epsilon())
  { return std::abs(a - b) < epsilon; }
  
  // 判断两个值是否不相等
  static bool neq(const T& a, const T& b, const T& epsilon = SGLimits<T>::epsilon())
  { return !eq(a, b, epsilon); }
};

#endif
